/*
    Copyright (C) 2001  bjk <bjk@arbornet.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>
#include <grp.h>
#include "common.h"

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

static int escapes(const char *str)
{
    static int c;

    if (str[0] != '\\')
	return str[0];

    switch (*++str) {
	case 't': c = '\t'; break;
	case 'n': c = '\n'; break;
	case '\\': c = '\\'; break;
	case 'v': c = '\v'; break;
	case 'b': c = '\b'; break;
	case 'f': c = '\f'; break;
	case 'r': c = '\r'; break;
	case '\'': c = '\''; break;
	default: c = 0; break;
    }

    return c;
}

static void usage(char *me)
{
    extern unsigned int delimchar;

    fprintf(stderr, "Usage: %s [-vhV] [-j tf] [-D c] [options] "
	    "-f <filename> ... | user ... | -\n", me);
    fprintf(stderr, "  Password/Group file info:\n");
    fprintf(stderr, "\t-l\tlogin name\t\t");
    fprintf(stderr, "\t-p\tencrypted password\n");
    fprintf(stderr, "\t-u\tuser id (uid)\t\t");
    fprintf(stderr, "\t-g\tgroup id (gid)\n");
    fprintf(stderr, "\t-c\tpassword change time\t");
    fprintf(stderr, "\t-k\tpassword expire time\n");
    fprintf(stderr, "\t-d\thome directory\t\t");
    fprintf(stderr, "\t-m\thome directory mode\n");
    fprintf(stderr, "\t-s\tlogin shell\n");
    fprintf(stderr, "\t-i opts\tgecos information "
	    "([n]ame,[o]ffice,[w]phone,[h]phone),[a]ll)\n\n");

    fprintf(stderr, "  Mail info:\t\t\t\t");
    fprintf(stderr, "  Login info:\n");
    fprintf(stderr, "\t-o\tforward address\t\t");
    fprintf(stderr, "\t-y\ttty\n");
    fprintf(stderr, "\t-a\tmail aliases\t\t");
    fprintf(stderr, "\t-e\tmessage status\n");
    fprintf(stderr, "\t-r\tlast read\t\t");
    fprintf(stderr, "\t-t\tlogin time stamp\n");
    fprintf(stderr, "\t-w\tnew mail\t\t");
    fprintf(stderr, "\t-b\tduration in minutes\n");
    fprintf(stderr, "\t-z\tfolder size\t\t");
    fprintf(stderr, "\t-x\thostname\n");
    fprintf(stderr, "\t\t\t\t\t\t-n\tlast login\n\n");
    

    fprintf(stderr, "  -D c\tSeparate output with specified character "
	    "('%c').\n", delimchar);
    fprintf(stderr, "  -j tf\tstrftime(3) time format ('%s').\n",
	    DEF_TIMEFORMAT);
    fprintf(stderr, "  -f\tTreat remaining arguments as filenames rather "
	    "than usernames.\n");
    fprintf(stderr, "  -v\tVerbose output; displays gids with group names, "
	    "and filenames.\n");
    fprintf(stderr, "  -h\tThis help text.\n");
    fprintf(stderr, "  -V\tVersion information.\n\n");
    fprintf(stderr, "Output key: %s=unknown, %s=none, %s=yes/on, %s=no/off\n",
	    UNKNOWN, NONE, ON, OFF);

    exit(-1);
}

static GECOS *get_gecos_opts(const char *args)
{
    int i = 0;
    GECOS *gecos;

    if (strchr(args, 'a') != NULL) {
	args = "nowh";
	get_gecos_opts(args);
    }

    if ((gecos = (GECOS *) malloc(sizeof(GECOS))) == NULL) {
	perror("malloc");
	exit(errno);
    }

    bzero(gecos, sizeof(GECOS));

    while (*args) {
	switch (*args++) {
	    case 'n':
	    case 'o':
	    case 'w':
	    case 'h':
		gecos->gecos_opts[i++] = *args;
		break;
	    default:
		return NULL;
	}
    }

    return gecos;
}

static int command_line_opts(int myargc, char **myargv, const char *opts)
{
    int opt;
    int mindex = 0, pindex = 0, lindex = 0, i = 0;
    static int run;
    extern unsigned int usefile, passwd_info, delimchar;
    extern unsigned int mail_info, login_info;
    extern int passwd_opts[], mail_opts[], optspec[], login_opts[];
    extern char *gecosopts;

    optind = 1;

    while ((opt = getopt(myargc, myargv, opts)) != EOF) {
	if (!run)
	    optspec[i++] = opt;

	switch (opt) {
	    case 'D':
		      if (optarg[0] != '\\' && strlen(optarg) > 1)
			  return 1;

		      if ((delimchar = escapes(optarg)) == 0)
			  return 1;
		      
		      break;
	    case 'j':
		      strncpy(tf, optarg, sizeof(tf)); break;
	    case 'h': return 1; break;
	    case 'V':
		      fprintf(stdout, "%s\n%s\n", VERSION, COPYRIGHT);
		      exit(0);
		      break;
	    case 'f': usefile = 1; break;
	    case 'v': verbose = 1; break;
	    case 'i': gecosopts = optarg;
	    case 'l':
	    case 's':
	    case 'u':
	    case 'd':
	    case 'm':
	    case 'p':
	    case 'c':
	    case 'k':
	    case 'g':
		      passwd_opts[pindex++] = opt;
		      passwd_info = 1;
		      break;
	    case 'z':
	    case 'o':
	    case 'r':
	    case 'a':
	    case 'w':
		      mail_opts[mindex++] = opt;
		      mail_info = 1;
		      break;
	    case 'y':
	    case 'e':
	    case 'b':
	    case 'x':
	    case 'n':
	    case 't':
		      login_opts[lindex++] = opt;
		      login_info = 1;
		      break;
	    case '?':
	    default: return 1; break;
	}
    }

    run++;
    return 0;
}

static void junction(INFO *info, const char *arg)
{
    extern unsigned int usefile, mail_info, passwd_info, login_info;
    char filename[NAME_MAX];
    struct stat mystat;
    
    setpwent();

#ifdef HAVE_GETSPNAM
    if (amroot)
	setspent();
#endif

    if (usefile) {
	strncpy(filename, arg, sizeof(filename));

        if ((lstat(filename, &mystat)) == -1) {
	    perror(filename);
	    return;
	}

        if ((info->passwd = getpwuid(mystat.st_uid)) == NULL) {
	    perror("no such user");
	    return;
        }

#ifdef HAVE_GETSPNAM
	if (amroot) {
	    if ((info->spwd = getspnam(info->passwd->pw_name)) == NULL) {
		fprintf(stderr, "getspnam(): unknown error\n");
	        return;
	    }
	} 
#endif
    }
    else {
        if ((info->passwd = getpwnam(arg)) == NULL) {
	    fprintf(stderr, "%s: no such user\n", arg);
	    return;
	} 

#ifdef HAVE_GETSPNAM
	if (amroot) {
	    if ((info->spwd = getspnam(arg)) == NULL) {
		fprintf(stderr, "getspnam(): unknown error\n");
	        return;
	    }
	} 
#endif
    }

    /* fill the info structure */
    if (passwd_info) 
	passwdinfo(info);

    if (mail_info)
	mailinfo(info, arg);

    if (login_info)
	logininfo(info, arg);

    outputinfo(info, arg);
    return;
}

static void cleanup(INFO *info)
{
    GROUP *gcur, *gnext;

    if (info->gecos)
	free(info->gecos);

    if (info->login)
	free(info->login);

    if (info->mail)
	free(info->mail);

    if (info->group) {
	for (gcur = info->group; gcur; gcur = gnext) {
	    gnext = gcur->next;
	    free(gcur);
	}
    }

    free(info);
    return;
}

int main(int argc, char *argv[])
{
    char progname[NAME_MAX];
    extern unsigned int is_stdin, delimchar;
    extern char *gecosopts;
    char *opts = "nkcj:aD:hmvplsdi:Vugfozrweytbx";

    delimchar = DEFDELIM;
    strncpy(tf, DEF_TIMEFORMAT, sizeof(tf));
    strncpy(progname, argv[0], sizeof(progname));

    if (command_line_opts(argc, argv, opts) != 0)
	usage(progname);

    if (argc == optind || optind < 2)
	usage(progname);

#ifdef HAVE_SETPASSENT
    if (!setpassent(1))
	fprintf(stderr, "cannot keep passwd file open (will be slower)\n");
#endif

#ifdef HAVE_GETSPNAM
    if (getuid() == 0)
	amroot = 1;
#endif

    while (optind < argc) {
	char *arg = argv[optind];
	char *buf;
	INFO *info;

	if (arg[0] == '-' && arg[1] == 0)
	    is_stdin = 1;

        if ((info = (INFO *) malloc(sizeof(INFO))) == NULL) {
            perror("malloc");
            exit(errno);
        }

        bzero(info, sizeof(INFO));

        if (gecosopts) {
            if ((info->gecos = get_gecos_opts(gecosopts)) == NULL)
                usage(progname);
        }
	
	if (is_stdin) {
	    if ((buf = (char *) malloc(LINE_MAX)) == NULL) {
		perror("malloc");
		exit(errno);
	    }

	    while ((buf = fgets(buf, LINE_MAX, stdin)) != NULL) {
        	if ((info = (INFO *) malloc(sizeof(INFO))) == NULL) {
                    perror("malloc");
                    exit(errno);
                }

                bzero(info, sizeof(INFO));
		arg = stripstr(buf, "\n");

		junction(info, arg);
		cleanup(info);
	    }

	    free(buf);
	    break;
	}

	junction(info, arg);
	cleanup(info);
	optind++;
    }

    if (utmpfd)
	close(utmpfd);

    if (lastfp)
	fclose(lastfp);

    if (aliasfp)
	fclose(aliasfp);

    if (passwd_info)
	endpwent();

#ifdef HAVE_GETSPNAM
    if (amroot) {
        if (passwd_info)
	    endspent();
    }
#endif

    if (group_info)
	endgrent();

    exit(0);
}
